from datetime import datetime, timezone
from unittest.mock import call, patch

from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus
from arkindex.ponos.models import Farm
from arkindex.process.models import (
    CorpusWorkerVersion,
    FeatureUsage,
    ProcessMode,
    Worker,
    WorkerConfiguration,
    WorkerType,
    WorkerVersion,
    WorkerVersionState,
)
from arkindex.project.tests import FixtureAPITestCase
from arkindex.training.models import Model, ModelVersionState
from arkindex.users.models import Role, User


class TestWorkersWorkerVersions(FixtureAPITestCase):
    """
    Test workers and worker versions endpoints
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.worker_type_dla = WorkerType.objects.get(slug="dla")
        cls.worker_reco = Worker.objects.get(slug="reco")
        cls.worker_dla = Worker.objects.get(slug="dla")
        cls.worker_gpu = Worker.objects.get(slug="worker-gpu")
        cls.worker_generic = Worker.objects.get(slug="generic")
        cls.worker_file_import = Worker.objects.get(slug="file_import")
        cls.worker_custom = Worker.objects.get(slug="custom")
        cls.init_worker = Worker.objects.get(slug="initialisation")
        cls.worker_custom.memberships.create(user=cls.user, level=Role.Contributor.value)

        cls.model = Model.objects.create(
            name="Generic model",
            public=False,
        )
        cls.model_version = cls.model.versions.create(
            state=ModelVersionState.Available,
            hash="42",
            archive_hash="42",
            size=1337,
        )

        cls.farm = Farm.objects.first()
        process = cls.corpus.processes.create(
            mode=ProcessMode.Workers,
            creator=cls.user,
            farm=cls.farm,
            name="test_process"
        )
        process.run()
        cls.task = process.tasks.get()

    def setUp(self):
        super().setUp()
        self.version_1 = WorkerVersion.objects.get(worker=self.worker_reco)
        self.version_2 = self.worker_dla.versions.get()
        self.version_custom = self.worker_custom.versions.get()

    # Tests on get_query_set for WorkerList
    def test_workers_list_requires_login(self):
        response = self.client.get(reverse("api:workers-list"))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_workers_list(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.get(reverse("api:workers-list"))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 7,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [
                {
                    "id": str(self.worker_custom.id),
                    "repository_url": None,
                    "name": "Custom worker",
                    "description": "",
                    "slug": "custom",
                    "type": "custom",
                    "archived": False,
                },
                {
                    "id": str(self.worker_dla.id),
                    "repository_url": None,
                    "name": "Document layout analyser",
                    "description": "",
                    "slug": "dla",
                    "type": "dla",
                    "archived": False,
                },
                {
                    "id": str(self.init_worker.id),
                    "repository_url": None,
                    "name": "Elements Initialisation Worker",
                    "description": "",
                    "slug": "initialisation",
                    "type": "init",
                    "archived": False,
                },
                {
                    "id": str(self.worker_file_import.id),
                    "repository_url": None,
                    "name": "File import",
                    "description": "",
                    "slug": "file_import",
                    "type": "import",
                    "archived": False,
                },
                {
                    "id": str(self.worker_generic.id),
                    "repository_url": None,
                    "name": "Generic worker with a Model",
                    "description": "",
                    "slug": "generic",
                    "type": "recognizer",
                    "archived": False,
                },
                {
                    "id": str(self.worker_reco.id),
                    "repository_url": None,
                    "name": "Recognizer",
                    "description": "",
                    "slug": "reco",
                    "type": "recognizer",
                    "archived": False,
                },
                {
                    "id": str(self.worker_gpu.id),
                    "repository_url": None,
                    "name": "Worker requiring a GPU",
                    "description": "",
                    "slug": "worker-gpu",
                    "type": "worker",
                    "archived": False,
                },
            ]
        })

    def test_workers_list_compatible_model(self):
        """
        User is able to filter workers by a compatible model.
        """
        self.model.memberships.create(user=self.user, level=Role.Guest.value)
        self.worker_generic.models.set([self.model])
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.get(
                reverse("api:workers-list"),
                {"compatible_model": str(self.model.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 1,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [
                {
                    "id": str(self.worker_generic.id),
                    "repository_url": None,
                    "name": "Generic worker with a Model",
                    "description": "",
                    "slug": "generic",
                    "type": "recognizer",
                    "archived": False,
                },
            ]
        })

    def test_workers_list_compatible_model_public(self):
        public_model = Model.objects.create(name="Public", public=True)
        self.worker_reco.models.set([public_model])
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.get(
                reverse("api:workers-list"),
                {"compatible_model": str(public_model.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.json(), {
            "count": 1,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [{
                "id": str(self.worker_reco.id),
                "repository_url": None,
                "name": "Recognizer",
                "description": "",
                "slug": "reco",
                "type": "recognizer",
                "archived": False,
            }],
        })

    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    def test_worker_list_compatible_model_private(self, filter_rights_mock):
        private_model = Model.objects.create(name="Private")
        self.worker_reco.models.set([private_model])
        self.client.force_login(self.user)
        filter_rights_mock.side_effect = [
            Worker.objects.all(),
            Model.objects.none(),
        ]

        with self.assertNumQueries(2):
            response = self.client.get(
                reverse("api:workers-list"),
                {"compatible_model": str(private_model.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "compatible_model": ["This model does not exist or you don't have access to it"]
        })

        self.assertListEqual(filter_rights_mock.call_args_list, [
            call(self.user, Worker, Role.Contributor.value),
            call(self.user, Model, Role.Guest.value),
        ])

    def test_workers_list_filter_invalid_query_params(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.get(
                reverse("api:workers-list"),
                {"compatible_model": "A"},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "compatible_model": ["Invalid UUID"],
        })

    def test_workers_list_filter_type_slug(self):
        """
        User is able to filter workers by worker type slug
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:workers-list"), {"type": "dla"})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 1,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [
                {
                    "id": str(self.worker_dla.id),
                    "repository_url": None,
                    "name": "Document layout analyser",
                    "description": "",
                    "slug": "dla",
                    "type": "dla",
                    "archived": False,
                }
            ]
        })

    def test_workers_list_filter_archived(self):
        self.client.force_login(self.user)
        self.worker_dla.archived = datetime.now(timezone.utc)
        self.worker_dla.save()

        with self.assertNumQueries(4):
            response = self.client.get(reverse("api:workers-list"), {"archived": True})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 1,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [
                {
                    "id": str(self.worker_dla.id),
                    "repository_url": None,
                    "name": "Document layout analyser",
                    "description": "",
                    "slug": "dla",
                    "type": "dla",
                    "archived": True,
                }
            ]
        })

    def test_workers_list_filter_unarchived(self):
        self.client.force_login(self.user)
        Worker.objects.all().update(archived=datetime.now(timezone.utc))
        self.worker_dla.archived = None
        self.worker_dla.save()

        with self.assertNumQueries(4):
            response = self.client.get(reverse("api:workers-list"), {"archived": False})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 1,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [
                {
                    "id": str(self.worker_dla.id),
                    "repository_url": None,
                    "name": "Document layout analyser",
                    "description": "",
                    "slug": "dla",
                    "type": "dla",
                    "archived": False,
                }
            ]
        })

    def test_workers_list_filter_invalid_type_slug(self):
        """
        Raise when trying to filter workers with an invalid type slug
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:workers-list"), {"type": "invalid-slug"})
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"type": ["No registered worker type with that slug."]})

    def test_workers_list_filter_type_id(self):
        """
        User is able to filter workers by worker type id
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.get(reverse("api:workers-list"), {"type": self.worker_type_dla.id})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 1,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [
                {
                    "id": str(self.worker_dla.id),
                    "repository_url": None,
                    "name": "Document layout analyser",
                    "description": "",
                    "slug": "dla",
                    "type": "dla",
                    "archived": False,
                }
            ]
        })

    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    def test_workers_list_requires_contributor(self, filter_rights_mock):
        """
        User is not able to list workers with a guest access
        """
        self.client.force_login(self.user)
        filter_rights_mock.side_effect = [
            Worker.objects.none(),
        ]

        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:workers-list"))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 0,
            "next": None,
            "number": 1,
            "previous": None,
            "results": []
        })

        self.assertListEqual(filter_rights_mock.call_args_list, [
            call(self.user, Worker, Role.Contributor.value),
        ])

    def test_workers_list_name_filter(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.get(reverse("api:workers-list"), {"name": "layout"})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 1,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [
                {
                    "id": str(self.worker_dla.id),
                    "repository_url": None,
                    "name": "Document layout analyser",
                    "description": "",
                    "slug": "dla",
                    "type": "dla",
                    "archived": False,
                }
            ]
        })

    def test_workers_retrieve_required_login(self):
        response = self.client.get(reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_workers_retrieve(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_custom.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "id": str(self.worker_custom.id),
            "repository_url": None,
            "name": "Custom worker",
            "description": "",
            "slug": "custom",
            "type": "custom",
            "archived": False,
        })

    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    def test_workers_retrieve_no_rights(self, filter_rights_mock):
        filter_rights_mock.side_effect = [
            Worker.objects.none(),
        ]
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(response.json(), {"detail": "No Worker matches the given query."})

        self.assertListEqual(filter_rights_mock.call_args_list, [
            call(self.user, Worker, Role.Contributor.value),
        ])

    def test_worker_create_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(reverse("api:workers-list"))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_worker_create_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.post(reverse("api:workers-list"))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    def test_worker_create_required_fields(self):
        self.user.can_create_worker_version = True
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.post(reverse("api:workers-list"))
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "name": ["This field is required."],
            "slug": ["This field is required."],
            "type": ["This field is required."]
        })

    def test_worker_create_user_requires_permission(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.post(
                reverse("api:workers-list"),
                data={
                    "name": "Worker post",
                    "slug": "worker_post",
                    "type": self.worker_type_dla.slug,
                    "repository_url": "https://gitlab.com/NERV/eva/"
                }
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to create workers or worker versions."})

    def test_worker_create_user(self):
        """
        A user can create a worker that is not linked to a repository.
        The user has an admin access to that worker.
        """
        self.user.can_create_worker_version = True
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.post(
                reverse("api:workers-list"),
                data={
                    "name": "Worker post",
                    "slug": "worker_post",
                    "type": self.worker_type_dla.slug,
                    "repository_url": "https://gitlab.com/NERV/eva/"
                }
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        worker = Worker.objects.get(id=data["id"])
        self.assertEqual(worker.type, self.worker_type_dla)
        self.assertQuerySetEqual(worker.memberships.values_list("user_id", "level"), [
            (self.user.id, Role.Admin.value)
        ])
        self.assertDictEqual(data, {
            "id": str(worker.id),
            "name": "Worker post",
            "description": "",
            "repository_url": "https://gitlab.com/NERV/eva/",
            "slug": "worker_post",
            "type": "dla",
            "archived": False,
        })

    @patch("arkindex.users.utils.get_max_level", return_value=Role.Guest.value)
    def test_worker_create_user_existing_no_execution_access(self, get_max_level_mock):
        """
        A user cannot retrieve a worker without an execution access in case the slug already exists.
        No worker type is created.
        """
        self.worker_custom.memberships.filter(user=self.user).update(level=Role.Guest.value)
        self.user.can_create_worker_version = True
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:workers-list"),
                data={
                    "name": "Worker post",
                    "slug": self.worker_custom.slug,
                    "type": "new_type"
                }
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertFalse(WorkerType.objects.filter(slug="new_type").exists())
        self.assertQuerySetEqual(self.worker_custom.memberships.values_list("user_id", "level"), [
            (self.user.id, Role.Guest.value)
        ])
        self.assertDictEqual(response.json(), {
            "repository_url": ["You cannot use this value."],
            "slug": ["You cannot use this value."],
        })
        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user, self.worker_custom))

    def test_worker_create_task_return_existing_worker(self):
        """
        Creation with an existing slug and no repo URL returns the existing worker with a HTTP status 200
        """
        test_worker = Worker.objects.create(
            name="Prog Knife",
            slug="prog_knife",
            type=self.worker_reco.type
        )
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:workers-list"),
                data={
                    "name": "Progressive Knife",
                    "slug": "prog_knife",
                    "type": self.worker_type_dla.slug,
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(data["id"], str(test_worker.id))
        self.assertEqual(data["name"], "Prog Knife")
        self.assertEqual(data["slug"], "prog_knife")
        self.assertEqual(data["type"], "recognizer")
        self.assertEqual(data["repository_url"], None)
        self.assertEqual(test_worker.repository_url, None)

    def test_worker_create_existing_with_repository_url(self):
        """
        Creation with an existing slug and repo URL returns the existing worker with a HTTP status 200
        """
        test_worker = Worker.objects.create(
            name="Prog Knife",
            slug="prog_knife",
            type=self.worker_reco.type,
            repository_url="https://gitlab.nerv.co.jp/eva:eva-01",
        )
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:workers-list"),
                data={
                    "name": "Progressive Knife",
                    "slug": "prog_knife",
                    "type": self.worker_type_dla.slug,
                    "repository_url": "https://gitlab.nerv.co.jp/eva:eva-01",
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(data["id"], str(test_worker.id))
        self.assertEqual(data["name"], "Prog Knife")
        self.assertEqual(data["slug"], "prog_knife")
        self.assertEqual(data["type"], "recognizer")
        self.assertEqual(data["repository_url"], "https://gitlab.nerv.co.jp/eva:eva-01")

    def test_worker_create_task_new_type(self):
        """
        Ponos task can create a worker with a non pre-existing worker type
        """
        new_worker_type_slug = "newType"
        self.assertFalse(WorkerType.objects.filter(slug=new_worker_type_slug).exists())

        with self.assertNumQueries(11):
            response = self.client.post(
                reverse("api:workers-list"),
                data={
                    "name": "Worker post",
                    "slug": "worker_post",
                    "type": "newType",
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        self.assertTrue(WorkerType.objects.filter(slug=new_worker_type_slug).exists())
        self.assertNotEqual(data["id"], str(self.worker_reco.id))
        self.assertEqual(data["name"], "Worker post")
        self.assertEqual(data["slug"], "worker_post")
        self.assertEqual(data["type"], "newType")
        self.assertEqual(data["repository_url"], None)

    def test_update_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.put(
                reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                {
                    "name": "New name",
                    "slug": "new_slug",
                    "description": "New description",
                    "type": "dla",
                    "archived": False,
                },
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_update_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.put(
                reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                {
                    "name": "New name",
                    "slug": "new_slug",
                    "description": "New description",
                    "type": "dla",
                    "archived": False,
                },
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    @patch("arkindex.users.utils.get_max_level", return_value=Role.Contributor.value)
    def test_update_requires_admin(self, max_level_mock, filter_rights_mock):
        self.worker_reco.memberships.update_or_create(user=self.user, defaults={"level": Role.Contributor.value})
        self.user.can_manage_workers = True
        self.user.save()
        self.client.force_login(self.user)
        filter_rights_mock.side_effect = [
            Worker.objects.all(),
        ]

        with self.assertNumQueries(3):
            response = self.client.put(
                reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                {
                    "name": "New name",
                    "slug": "new_slug",
                    "description": "New description",
                    "type": "dla",
                    "archived": False,
                },
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {"detail": "You do not have an admin access to this worker."})
        self.assertListEqual(filter_rights_mock.call_args_list, [
            call(self.user, Worker, Role.Contributor.value),
        ])
        self.assertEqual(max_level_mock.call_count, 1)
        self.assertEqual(max_level_mock.call_args, call(self.user, self.worker_reco))

    def test_update_requires_permission(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.put(
                reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                {
                    "name": "New name",
                    "slug": "new_slug",
                    "description": "New description",
                    "type": "dla",
                    "archived": False,
                },
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {"detail": "You do not have permission to manage workers."})

    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    def test_update(self, filter_rights_mock):
        self.user.can_manage_workers = True
        self.user.save()
        self.client.force_login(self.user)
        filter_rights_mock.side_effect = [
            Worker.objects.all(),
        ]

        with self.assertNumQueries(7):
            response = self.client.put(
                reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                {
                    "name": "New name",
                    "slug": "new_slug",
                    "description": "New description",
                    "type": "dla",
                    "archived": False,
                    "repository_url": "https://gitlab.com/NERV/eva"
                },
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.json(), {
            "id": str(self.worker_reco.id),
            "name": "New name",
            "slug": "new_slug",
            "description": "New description",
            "type": "dla",
            "repository_url": "https://gitlab.com/NERV/eva",
            "archived": False,
        })

        self.worker_reco.refresh_from_db()
        self.assertEqual(self.worker_reco.name, "New name")
        self.assertEqual(self.worker_reco.slug, "new_slug")
        self.assertEqual(self.worker_reco.description, "New description")
        self.assertEqual(self.worker_reco.type, self.worker_type_dla)
        self.assertEqual(self.worker_reco.repository_url, "https://gitlab.com/NERV/eva")
        self.assertListEqual(filter_rights_mock.call_args_list, [
            call(self.user, Worker, Role.Contributor.value),
        ])

    def test_update_unique_slug(self):
        self.user.can_manage_workers = True
        self.user.save()
        self.client.force_login(self.user)

        self.assertEqual(self.worker_reco.repository_url, None)
        with self.assertNumQueries(4):
            response = self.client.put(
                reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                {
                    "name": "New name",
                    "slug": "dla",
                    "description": "New description",
                    "type": "recognizer",
                    "archived": False,
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "non_field_errors": ["A worker with the same repository URL and slug already exists."],
        })

    def test_update_archived(self):
        self.user.can_manage_workers = True
        self.user.save()
        self.client.force_login(self.user)
        self.worker_reco.memberships.update_or_create(user=self.user, defaults={"level": Role.Admin.value})
        self.assertTrue(self.worker_reco.is_archivable(self.user))

        cases = [
            (datetime.now(timezone.utc), False),
            (None, True),
        ]

        for current_value, new_value in cases:
            with self.subTest(current_value=current_value, new_value=new_value):
                self.worker_reco.archived = current_value
                self.worker_reco.save()

                with self.assertNumQueries(7):
                    response = self.client.put(
                        reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                        {
                            "name": "New name",
                            "slug": "new_slug",
                            "description": "New description",
                            "type": "dla",
                            "archived": new_value,
                        },
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)

                self.assertEqual(response.json(), {
                    "id": str(self.worker_reco.id),
                    "name": "New name",
                    "slug": "new_slug",
                    "description": "New description",
                    "type": "dla",
                    "repository_url": None,
                    "archived": new_value,
                })

    def test_update_new_type(self):
        """
        A new type is automatically created if necessary
        """
        self.user.can_manage_workers = True
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(10):
            response = self.client.patch(
                reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                {
                    "name": "New name",
                    "slug": "new_slug",
                    "description": "New description",
                    "type": "new_type",
                    "archived": False,
                },
            )

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json(), {
            "archived": False,
            "description": "New description",
            "id": str(self.worker_reco.id),
            "name": "New name",
            "repository_url": None,
            "slug": "new_slug",
            "type": "new_type",
        })
        self.worker_reco.refresh_from_db()
        self.assertEqual(self.worker_reco.type.slug, "new_type")
        self.assertEqual(self.worker_reco.type.display_name, "New_type")

    def test_partial_update_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.patch(
                reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                {
                    "description": "New description",
                },
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_partial_update_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.patch(
                reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                {
                    "description": "New description",
                },
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    @patch("arkindex.users.utils.get_max_level", return_value=Role.Contributor.value)
    def test_partial_update_requires_admin(self, max_level_mock, filter_rights_mock):
        self.worker_reco.memberships.update_or_create(user=self.user, defaults={"level": Role.Contributor.value})
        self.user.can_manage_workers = True
        self.user.save()
        self.client.force_login(self.user)
        filter_rights_mock.side_effect = [
            Worker.objects.all(),
        ]

        with self.assertNumQueries(3):
            response = self.client.patch(
                reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                {
                    "description": "New description",
                },
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {"detail": "You do not have an admin access to this worker."})

        self.assertListEqual(filter_rights_mock.call_args_list, [
            call(self.user, Worker, Role.Contributor.value),
        ])
        self.assertEqual(max_level_mock.call_count, 1)
        self.assertEqual(max_level_mock.call_args, call(self.user, self.worker_reco))

    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    def test_partial_update(self, filter_rights_mock):
        self.worker_reco.memberships.update_or_create(user=self.user, defaults={"level": Role.Admin.value})
        self.user.can_manage_workers = True
        self.user.save()
        self.client.force_login(self.user)
        filter_rights_mock.side_effect = [
            Worker.objects.all()
        ]

        with self.assertNumQueries(6):
            response = self.client.patch(
                reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                {
                    "description": "New description",
                    "type": "dla",
                },
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.json(), {
            "id": str(self.worker_reco.id),
            "name": "Recognizer",
            "slug": "reco",
            "description": "New description",
            "type": "dla",
            "repository_url": None,
            "archived": False,
        })

        self.worker_reco.refresh_from_db()
        self.assertEqual(self.worker_reco.name, "Recognizer")
        self.assertEqual(self.worker_reco.slug, "reco")
        self.assertEqual(self.worker_reco.description, "New description")
        self.assertEqual(self.worker_reco.type, self.worker_type_dla)

        self.assertListEqual(filter_rights_mock.call_args_list, [
            call(self.user, Worker, Role.Contributor.value),
        ])

    def test_partial_update_unique_slug(self):
        self.user.can_manage_workers = True
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.patch(
                reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                {
                    "slug": "dla",
                },
            )

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "non_field_errors": ["A worker with the same repository URL and slug already exists."],
        })

    def test_partial_update_archived(self):
        self.user.can_manage_workers = True
        self.user.save()
        self.client.force_login(self.user)
        self.worker_reco.memberships.update_or_create(user=self.user, defaults={"level": Role.Admin.value})
        self.assertTrue(self.worker_reco.is_archivable(self.user))

        cases = [
            (datetime.now(timezone.utc), False),
            (None, True),
        ]

        for current_value, new_value in cases:
            with self.subTest(current_value=current_value, new_value=new_value):
                self.worker_reco.archived = current_value
                self.worker_reco.save()

                with self.assertNumQueries(4):
                    response = self.client.patch(
                        reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                        {
                            "archived": new_value,
                        },
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)

                self.assertEqual(response.json(), {
                    "id": str(self.worker_reco.id),
                    "name": "Recognizer",
                    "slug": "reco",
                    "description": "",
                    "type": "recognizer",
                    "repository_url": None,
                    "archived": new_value,
                })

    def test_partial_update_new_type(self):
        """
        A new type is automatically created if necessary
        """
        self.user.can_manage_workers = True
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(9):
            response = self.client.patch(
                reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                {"type": "new_type"},
            )

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json(), {
            "archived": False,
            "description": "",
            "id": str(self.worker_reco.id),
            "name": "Recognizer",
            "repository_url": None,
            "slug": "reco",
            "type": "new_type",
        })
        self.worker_reco.refresh_from_db()
        self.assertEqual(self.worker_reco.type.slug, "new_type")
        self.assertEqual(self.worker_reco.type.display_name, "New_type")

    # Tests on get_query_set for WorkerVersionList
    def test_versions_list_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_versions_list(self):
        self.client.force_login(self.user)
        last_version = self.worker_reco.versions.create(
            configuration={"a": "bc"},
            state=WorkerVersionState.Error,
            revision_url="https://gitlab.com/NERV/eva/commit/63e377e7f88c743d8428fc4e4eaedfc1c9356754"
        )
        last_version.created = "2050-09-09T09:09:09.090909Z"
        last_version.save()

        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 2,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [
                {
                    "id": str(last_version.id),
                    "configuration": {"a": "bc"},
                    "docker_image_iid": last_version.docker_image_iid,
                    "state": "error",
                    "gpu_usage": FeatureUsage.Disabled.value,
                    "model_usage": FeatureUsage.Disabled.value,
                    "worker": {
                        "id": str(self.worker_reco.id),
                        "name": self.worker_reco.name,
                        "type": self.worker_reco.type.slug,
                        "slug": self.worker_reco.slug,
                        "description": self.worker_reco.description,
                        "archived": bool(self.worker_reco.archived),
                        "repository_url": self.worker_reco.repository_url,
                    },
                    "version": None,
                    "tag": None,
                    "branch": None,
                    "revision_url": "https://gitlab.com/NERV/eva/commit/63e377e7f88c743d8428fc4e4eaedfc1c9356754",
                    "created": "2050-09-09T09:09:09.090909Z",
                },
                {
                    "id": str(self.version_1.id),
                    "configuration": {"test": 42},
                    "docker_image_iid": self.version_1.docker_image_iid,
                    "state": "available",
                    "gpu_usage": FeatureUsage.Disabled.value,
                    "model_usage": FeatureUsage.Disabled.value,
                    "worker": {
                        "id": str(self.worker_reco.id),
                        "name": self.worker_reco.name,
                        "type": self.worker_reco.type.slug,
                        "slug": self.worker_reco.slug,
                        "description": self.worker_reco.description,
                        "archived": bool(self.worker_reco.archived),
                        "repository_url": self.worker_reco.repository_url,
                    },
                    "version": 1,
                    "tag": None,
                    "branch": None,
                    "revision_url": None,
                    "created": self.version_1.created.isoformat().replace("+00:00", "Z"),
                }
            ]
        })

    @patch("arkindex.users.utils.get_max_level", return_value=Role.Guest.value)
    def test_versions_list_requires_contributor(self, max_level_mock):
        """
        User is not able to list worker versions with a guest access
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {
            "detail": "You do not have an execution access to this worker.",
        })
        self.assertEqual(max_level_mock.call_count, 1)
        self.assertEqual(max_level_mock.call_args_list, [
            call(self.user, self.worker_reco)
        ])

    def test_versions_list_public_worker(self):
        """
        Any authenticated user is able to list versions for public workers
        """
        user = User.objects.create_user("user42@test.fr", "abcd")
        user.verified_email = True
        user.save()
        self.worker_reco.public = True
        self.worker_reco.save()
        self.client.force_login(user)

        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(data["count"], 1)
        self.assertEqual(data["results"][0]["id"], str(self.worker_reco.versions.first().id))

    def test_versions_list_filter_worker(self):
        self.client.force_login(self.user)
        worker_2 = Worker.objects.create(
            name="Worker 2",
            slug="worker_2",
            type=self.worker_type_dla
        )
        version_2 = WorkerVersion.objects.create(
            worker=worker_2,
            version=1,
            configuration={"test": "test2"},
            gpu_usage=FeatureUsage.Disabled
        )

        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:worker-versions", kwargs={"pk": str(worker_2.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(len(data["results"]), 1)
        version = data["results"][0]
        self.assertEqual(version["id"], str(version_2.id))

    def test_versions_list_simple_mode(self):
        """
        With the mode attribute set to simple, worker versions with no tag are excluded
        """
        WorkerVersion.objects.bulk_create([
            WorkerVersion(
                worker=self.worker_reco,
                configuration={"test": "42"},
                tag=tag,
                version=i,
            ) for i, tag in enumerate((None, "v0.1", "test"), start=2)
        ])

        with self.subTest(mode="complete"), self.assertNumQueries(4):
            response = self.client.get(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                {"mode": "complete"},
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            data = response.json()
            self.assertEqual(data["count"], 4)
            self.assertCountEqual(
                [version["tag"] for version in data["results"]],
                ["v0.1", "test", None, None],
            )

        # Simple mode filters out versions without a tag
        with self.subTest(mode="simple"), self.assertNumQueries(4):
            response = self.client.get(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                {"mode": "simple"},
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            data = response.json()
            self.assertEqual(data["count"], 2)
            self.assertCountEqual(
                [version["tag"] for version in data["results"]],
                ["v0.1", "test"],
            )

    def test_create_version_non_existing_worker(self):
        with self.assertNumQueries(2):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": "12341234-1234-1234-1234-123412341234"}),
                data={"configuration": {"test": "test2"}},
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertDictEqual(response.json(), {"detail": "No Worker matches the given query."})

    def test_create_version_available_requires_docker_image(self):
        with self.assertNumQueries(2):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                data={
                    "configuration": {"test": "test2"},
                    "state": "available",
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "state": [
                "A worker version must have a `docker_image_iid` set when updated to available."
            ]
        })

    def test_create_version_invalid_docker_image(self):
        with self.assertNumQueries(2):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                data={
                    "docker_image_iid": "http://seele.co.jp/",
                    "configuration": {"test": "test2"},
                    "state": "available",
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "docker_image_iid": [
                "Repository name cannot contain a scheme (http://seele.co.jp/)"
            ]
        })

    def test_create_version_empty(self):
        self.user.can_create_worker_version = True
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"configuration": ["This field is required."]})

    def test_create_version_archived(self):
        self.worker_reco.archived = datetime.now(timezone.utc)
        self.worker_reco.save()

        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                data={
                    "configuration": {},
                    "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01"
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "worker": ["This worker is archived."],
        })

    def test_create_version_null_revision_url_requires_null_worker_repository_url(self):
        self.worker_reco.repository_url = "https://gitlab.com/NERV/eva"
        self.worker_reco.save()
        self.user.can_create_worker_version = True
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                data={"configuration": {"test": "test2"}, "model_usage": FeatureUsage.Required.value},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "revision_url": ["A revision url is required when creating a version for a worker linked to a repository."],
        })

    def test_create_version_null_revision_url_forbidden_task_auth(self):
        """
        Ponos Task auth cannot create a version on a worker that is not linked to a repository.
        """
        with self.assertNumQueries(2):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_custom.id)}),
                data={"configuration": {"test": "test2"}, "model_usage": FeatureUsage.Required.value},
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "revision_url": ["This field is required when creating a version through Task authentication."]
        })

    @patch("arkindex.users.utils.get_max_level", return_value=Role.Contributor.value)
    def test_create_version_user_auth_requires_admin_access_to_worker(self, max_level_mock):
        self.user.can_create_worker_version = True
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_custom.id)}),
                data={"configuration": {"test": "val"}, "model_usage": FeatureUsage.Required.value},
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {"detail": "You do not have an admin access to this worker."})

        self.assertEqual(max_level_mock.call_count, 1)
        self.assertEqual(max_level_mock.call_args, call(self.user, self.worker_custom))

    def test_create_version_user_auth_with_worker_repository_url_ok(self):
        self.user.can_create_worker_version = True
        self.user.save()
        self.client.force_login(self.user)
        self.worker_dla.repository_url = "https://gitlab.com/NERV/eva"
        self.worker_dla.save()

        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
                data={"revision_url": "https://gitlab.com/NERV/eva/commit/eva-01", "configuration": {"test": "test2"}, "model_usage": FeatureUsage.Required.value},
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        new_version = self.worker_dla.versions.get(id=data["id"])
        self.assertDictEqual(response.json(), {
            "id": str(new_version.id),
            "configuration": {"test": "test2"},
            "created": new_version.created.isoformat().replace("+00:00", "Z"),
            "docker_image_iid": None,
            "gpu_usage": FeatureUsage.Disabled.value,
            "model_usage": FeatureUsage.Required.value,
            "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01",
            "state": WorkerVersionState.Created.value,
            "version": None,
            "tag": None,
            "branch": None,
            "worker": {
                "id": str(self.worker_dla.id),
                "name": "Document layout analyser",
                "slug": "dla",
                "type": "dla",
                "description": "",
                "archived": False,
                "repository_url": "https://gitlab.com/NERV/eva",
            },
        })

    def test_create_version_null_revision_url(self):
        """
        A worker version can be created with no revision_url through user authentication
        """
        self.worker_custom.memberships.filter(user=self.user).update(level=Role.Admin.value)
        self.worker_custom.versions.create(version=41, configuration={})
        self.user.can_create_worker_version = True
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_custom.id)}),
                data={"configuration": {"test": "val"}, "model_usage": FeatureUsage.Required.value},
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        new_version = self.worker_custom.versions.get(id=data["id"])
        self.assertDictEqual(response.json(), {
            "id": str(new_version.id),
            "configuration": {"test": "val"},
            "created": new_version.created.isoformat().replace("+00:00", "Z"),
            "docker_image_iid": None,
            "gpu_usage": FeatureUsage.Disabled.value,
            "model_usage": FeatureUsage.Required.value,
            "revision_url": None,
            "state": WorkerVersionState.Created.value,
            "version": 42,
            "tag": None,
            "branch": None,
            "worker": {
                "id": str(self.worker_custom.id),
                "name": "Custom worker",
                "slug": "custom",
                "type": "custom",
                "description": "",
                "archived": False,
                "repository_url": None,
            },
        })

    def test_create_version_configuration_wrong_type(self):
        """
        Configuration body must be an object
        """
        self.user.can_create_worker_version = True
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                data={"configuration": "test"},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"configuration": ['Expected a dictionary of items but got type "str".']})

    def test_create_version_requires_permission(self):
        self.assertFalse(self.user.can_create_worker_version)
        self.client.force_login(self.user)
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
            data={
                "configuration": {"test": "test2"},
                "model_usage": FeatureUsage.Required.value,
                "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01",
            }
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {"detail": "You do not have permission to create workers or worker versions."})
        self.user.can_create_worker_version = True
        self.user.save()
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
            data={
                "configuration": {"test": "test2"},
                "model_usage": FeatureUsage.Required.value,
                "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01",
            }
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_create_version_task_auth(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
            data={
                "configuration": {"test": "test2"},
                "model_usage": FeatureUsage.Required.value,
                "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        self.assertNotEqual(data["id"], str(self.version_1.id))
        self.assertEqual(data["configuration"], {"test": "test2"})
        self.assertEqual(data["version"], None)
        self.assertEqual(data["revision_url"], "https://gitlab.com/NERV/eva/commit/eva-01")
        self.assertEqual(data["state"], "created")
        self.assertEqual(data["gpu_usage"], "disabled")
        self.assertEqual(data["model_usage"], FeatureUsage.Required.value)

    def test_create_version_duplicate_revision_url(self):
        self.version_1.revision_url = "https://gitlab.com/NERV/eva/commit/eva-01"
        self.version_1.version = None
        self.version_1.save()
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
            data={
                "configuration": {"test": "test2"},
                "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"revision_url": ["A version already exists for this worker with this revision_url."]})

    def test_create_version_with_tag(self):
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                data={
                    "configuration": {"test": "test2"},
                    "model_usage": FeatureUsage.Required.value,
                    "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01",
                    "tag": "eva-01",
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        self.assertNotEqual(data["id"], str(self.version_1.id))
        self.assertEqual(data["configuration"], {"test": "test2"})
        self.assertEqual(data["version"], None)
        self.assertEqual(data["revision_url"], "https://gitlab.com/NERV/eva/commit/eva-01")
        self.assertEqual(data["state"], "created")
        self.assertEqual(data["gpu_usage"], "disabled")
        self.assertEqual(data["model_usage"], FeatureUsage.Required.value)
        self.assertEqual(data["tag"], "eva-01")

    def test_create_version_unique_tag(self):
        self.version_1.tag = "eva-01"
        self.version_1.save()

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                data={"configuration": {"test": "test2"}, "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01", "tag": "eva-01"},
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"tag": ["A version already exists for this worker with this tag."]})

    def test_create_version_unassign_branch(self):
        self.version_1.branch = "operation-yashima"
        self.version_1.save()
        self.version_2.branch = "operation-yashima"
        self.version_2.save()

        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                data={
                    "configuration": {"test": "test2"},
                    "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01",
                    "branch": "operation-yashima",
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        new_version = WorkerVersion.objects.get(id=response.json()["id"])
        self.assertEqual(new_version.branch, "operation-yashima")

        self.version_1.refresh_from_db()
        self.assertIsNone(self.version_1.branch)

        self.version_2.refresh_from_db()
        self.assertEqual(self.version_2.branch, "operation-yashima")

    def test_create_version_empty_tag_branch(self):
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                data={
                    "configuration": {"test": "test2"},
                    "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01",
                    "tag": "",
                    "branch": "",
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "tag": ["This field may not be blank."],
            "branch": ["This field may not be blank."],
        })

    def test_create_version_wrong_gpu_usage(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
            data={"configuration": {"test": "test2"}, "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01", "gpu_usage": "not_supported"},
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_create_version_no_user_configuration_ok(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {"beep": "boop"},
                "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01",
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_create_version_valid_user_configuration(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "demo_integer": {"title": "Demo Integer", "type": "int", "required": True, "default": 1},
                        "demo_boolean": {"title": "Demo Boolean", "type": "bool", "required": False, "default": True},
                    }
                },
                "gpu_usage": "disabled",
                "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.json()["configuration"], {
            "user_configuration": {
                "demo_integer": {
                    "title": "Demo Integer",
                    "type": "int",
                    "required": True,
                    "default": 1
                },
                "demo_boolean": {
                    "title": "Demo Boolean",
                    "type": "bool",
                    "required": False,
                    "default": True
                }
            }
        })

    def test_create_version_valid_user_configuration_dict(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "demo_dict": {"title": "Demo Dict", "type": "dict", "required": True, "default": {"a": "b", "c": "d"}},
                    }
                },
                "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01",
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.json()["configuration"], {
            "user_configuration": {
                "demo_dict": {
                    "title": "Demo Dict",
                    "type": "dict",
                    "required": True,
                    "default": {"a": "b", "c": "d"}
                }
            }
        })

    def test_create_version_user_configuration_dict_strings_only(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "demo_dict": {"title": "Demo Dict", "type": "dict", "required": True, "default": {"a": ["12", "13"], "c": "d"}},
                    }
                },
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_create_version_valid_user_configuration_enum(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "demo_choice": {"title": "Decisions", "type": "enum", "required": True, "default": 1, "choices": [1, 2, 3]}
                    }
                },
                "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01",
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.json()["configuration"], {
            "user_configuration": {
                "demo_choice": {
                    "title": "Decisions",
                    "type": "enum",
                    "required": True,
                    "default": 1,
                    "choices": [1, 2, 3]
                }
            }
        })

    def test_create_version_valid_user_configuration_list(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "demo_list": {"title": "Demo List", "type": "list", "required": True, "subtype": "int", "default": [1, 2, 3, 4]},
                        "boolean_list": {"title": "It's a list of booleans", "type": "list", "required": False, "subtype": "bool", "default": [True, False, False]}
                    }
                },
                "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01",
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.json()["configuration"], {
            "user_configuration": {
                "demo_list": {
                    "title": "Demo List",
                    "type": "list",
                    "subtype": "int",
                    "required": True,
                    "default": [1, 2, 3, 4]
                },
                "boolean_list": {
                    "title": "It's a list of booleans",
                    "type": "list",
                    "subtype": "bool",
                    "required": False,
                    "default": [True, False, False]
                }
            }
        })

    def test_create_version_valid_user_configuration_model(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "demo_model": {"title": "Model for training", "type": "model", "required": True},
                        "other_model": {"title": "Model the second", "type": "model", "default": str(self.model.id)}
                    }
                },
                "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01",
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.json()["configuration"], {
            "user_configuration": {
                "demo_model": {
                    "title": "Model for training",
                    "type": "model",
                    "required": True
                },
                "other_model": {
                    "title": "Model the second",
                    "type": "model",
                    "default": str(self.model.id)
                }
            }
        })

    def test_create_version_invalid_user_configuration_list_requires_subtype(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "demo_list": {"title": "Demo List", "type": "list", "required": True, "default": [1, 2, 3, 4]},
                    }
                },
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [{
                    "demo_list": {
                        "subtype": ['The "subtype" field must be set for "list" type properties.']
                    }
                }]
            }
        })

    def test_create_version_invalid_user_configuration_list_wrong_default(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "demo_list": {"title": "Demo List", "type": "list", "required": True, "subtype": "int", "default": 12},
                    }
                },
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [{
                    "demo_list": {
                        "default": ["This is not a valid value for a field of type list."]
                    }
                }]
            }
        })

    def test_create_version_invalid_user_configuration_list_wrong_subtype(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "demo_list": {"title": "Demo List", "type": "list", "required": True, "subtype": "dict", "default": [1, 2, 3, 4]},
                    }
                },
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [{
                    "demo_list": {
                        "subtype": ["Subtype can only be int, float, bool or string."]
                    }
                }]
            }
        })

    def test_create_version_invalid_user_configuration_list_wrong_default_subtype(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "demo_list": {"title": "Demo List", "type": "list", "required": True, "subtype": "int", "default": [1, 2, "three", 4]},
                    }
                },
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [{
                    "demo_list": {
                        "default": ["All items in the default value must be of type int."]
                    }
                }]
            }
        })

    def test_create_version_invalid_user_configuration_not_list_choices(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "demo_choice": {"title": "Decisions", "type": "enum", "required": True, "default": 1, "choices": "eeee"}
                    }
                },
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [{
                    "demo_choice": {
                        "choices": ['Expected a list of items but got type "str".']
                    }
                }]
            }
        })

    def test_create_version_invalid_user_configuration_not_dict(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": "non"
                },
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"configuration": {"user_configuration": [['Expected a dictionary of items but got type "str".']]}})

    def test_create_version_invalid_user_configuration_item_not_dict(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "secrets": ["aaaaaa"]
                    }
                },
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"configuration": {
            "user_configuration": [{
                "secrets":
                {"__all__": ["User configuration field definitions should be of type dict, not list."]}
            }]
        }})

    def test_create_version_invalid_user_configuration_wrong_field_type(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "something": {
                            "title": "some thing",
                            "type": "uh oh",
                            "required": 2
                        }
                    }
                },
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [{
                    "something": {
                        "required": ["Must be a valid boolean."],
                        "type": ["Value is not of type UserConfigurationFieldType"]
                    }
                }]
            }
        })

    def test_create_version_invalid_user_configuration_wrong_default_type(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "one_float": {
                            "title": "a float",
                            "type": "float",
                            "default": "bonjour",
                            "required": True
                        }
                    }
                },
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [{
                    "one_float": {
                        "default": ["This is not a valid value for a field of type float."]
                    }
                }]
            }
        })

    def test_create_version_invalid_user_configuration_choices_no_enum(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "something": {
                            "title": "some thing",
                            "type": "int",
                            "required": False,
                            "choices": [1, 2, 3]
                        }
                    }
                },
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [{
                    "something": {
                        "choices": ['The "choices" field can only be set for an "enum" type property.']
                    }
                }]
            }
        })

    def test_create_version_invalid_user_configuration_missing_key(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "demo_integer": {"type": "int", "required": True, "default": 1}
                    }
                },
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.json(),
            {
                "configuration": {
                    "user_configuration": [{
                        "demo_integer": {
                            "title": ["This field is required."]
                        }
                    }]
                }
            }
        )

    def test_create_version_invalid_user_configuration_invalid_key(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "demo_integer": {
                            "title": "an integer",
                            "type": "int",
                            "required": True,
                            "default": 1,
                            "some_key": "oh no",
                        }
                    }
                },
                "gpu_usage": "disabled",
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [
                    {
                        "demo_integer": {
                            "some_key": ["Configurable properties can only be defined using the following keys: title, type, required, default, subtype, choices."]
                        }
                    }
                ]
            }
        })

    def test_create_version_invalid_user_configuration_default_value(self):
        cases = [
            ({"type": "int", "default": False}, "int"),
            ({"type": "int", "default": True}, "int"),
            ({"type": "float", "default": False}, "float"),
            ({"type": "float", "default": True}, "float"),
            ({"type": "bool", "default": 0}, "bool"),
            ({"type": "bool", "default": 1}, "bool"),
            ({"type": "string", "default": 1}, "string"),
            ({"type": "dict", "default": ["a", "b"]}, "dict"),
            ({"type": "model", "default": "gigi hadid"}, "model"),
            ({"type": "model", "default": False}, "model")
        ]
        for params, expected in cases:
            with self.subTest(**params):
                response = self.client.post(
                    reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
                    data={
                        "configuration": {
                            "user_configuration": {
                                "param": {"title": "param", **params}
                            }
                        },
                    },
                    HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
                )
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertEqual(response.json(), {"configuration": {"user_configuration": [{"param": {"default": [f"This is not a valid value for a field of type {expected}."]}}]}})

    def test_create_version_user_configuration_model_default_doesnt_exist(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "configuration": {
                    "user_configuration": {
                        "param": {"title": "Model to train", "type": "model", "default": "12341234-1234-1234-1234-123412341234"}
                    }
                },
            },
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"configuration": {"user_configuration": [{"param": {"default": [
            "Model 12341234-1234-1234-1234-123412341234 not found."
        ]}}]}})

    def test_retrieve_version_invalid_id(self):
        self.client.force_login(self.user)
        response = self.client.get(
            reverse("api:version-retrieve", kwargs={"pk": "12341234-1234-1234-1234-123412341234"})
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_retrieve_version_no_login(self):
        """
        Worker versions may be retrieved with no authentication in order
        to see the version on public resources
        """
        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:version-retrieve", kwargs={"pk": str(self.version_1.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertDictEqual(data, {
            "id": str(self.version_1.id),
            "configuration": {"test": 42},
            "docker_image_iid": self.version_1.docker_image_iid,
            "state": "available",
            "gpu_usage": "disabled",
            "model_usage": FeatureUsage.Disabled.value,
            "revision_url": None,
            "version": 1,
            "tag": None,
            "branch": None,
            "worker": {
                "id": str(self.worker_reco.id),
                "name": "Recognizer",
                "type": "recognizer",
                "slug": "reco",
                "description": "",
                "archived": False,
                "repository_url": None,
            },
            "created": "2020-02-02T01:23:45.678000Z",
        })

    def test_retrieve_version(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse("api:version-retrieve", kwargs={"pk": str(self.version_1.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertDictEqual(data, {
            "id": str(self.version_1.id),
            "configuration": {"test": 42},
            "docker_image_iid": self.version_1.docker_image_iid,
            "state": "available",
            "gpu_usage": "disabled",
            "model_usage": FeatureUsage.Disabled.value,
            "revision_url": None,
            "version": 1,
            "tag": None,
            "branch": None,
            "worker": {
                "id": str(self.worker_reco.id),
                "name": "Recognizer",
                "type": "recognizer",
                "slug": "reco",
                "description": "",
                "archived": False,
                "repository_url": None,
            },
            "created": "2020-02-02T01:23:45.678000Z",
        })

    def test_update_version_requires_login(self):
        response = self.client.patch(
            reverse("api:version-retrieve", kwargs={"pk": str(self.version_1.id)}),
            data={
                "configuration": {"test": "test2"},
                "docker_image_iid": "registry.nerv.co.jp/cruel/angel/thesis:latest",
                "state": "error",
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
            },
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_update_version_invalid_id(self):
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse("api:version-retrieve", kwargs={"pk": "12341234-1234-1234-1234-123412341234"}),
            data={
                "configuration": {"test": "test2"},
                "docker_image_iid": "registry.nerv.co.jp/cruel/angel/thesis:latest",
                "state": "error",
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
            },
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_update_worker_version_non_task(self):
        """
        Only Ponos tasks are allowed to update a worker version
        """
        docker_image_iid = self.version_1.docker_image_iid
        self.version_1.state = WorkerVersionState.Created
        self.version_1.docker_image_iid = None
        self.version_1.save()
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse("api:version-retrieve", kwargs={"pk": str(self.version_1.id)}),
            data={
                "configuration": {"test": "test2"},
                "docker_image_iid": str(docker_image_iid),
                "state": "error",
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
            },
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {"detail": "Only Ponos tasks are allowed to update a worker version."})

    def test_update_worker_version(self):
        """
        Update worker version artifact, configuration, state and tag
        Updating the branch should reset the branch on any other version on the same worker
        """
        self.version_1.state = WorkerVersionState.Created
        self.version_1.save()

        self.assertNotEqual(self.version_1.worker_id, self.version_2.worker_id)
        self.version_2.branch = "operation-yashima"
        self.version_2.save()

        branch_version = self.version_1.worker.versions.create(branch="operation-yashima", version=2)

        with self.assertNumQueries(6):
            response = self.client.patch(
                reverse("api:version-retrieve", kwargs={"pk": str(self.version_1.id)}),
                data={
                    "configuration": {"test": "test2"},
                    "docker_image_iid": "eva:unit-01",
                    "state": "error",
                    "tag": "eva-01",
                    "branch": "operation-yashima",
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(data["id"], str(self.version_1.id))

        self.version_1.refresh_from_db()
        self.assertEqual(self.version_1.configuration, {"test": "test2"})
        self.assertEqual(self.version_1.docker_image_iid, "eva:unit-01")
        self.assertEqual(self.version_1.state, WorkerVersionState.Error)
        self.assertEqual(self.version_1.gpu_usage, FeatureUsage.Disabled)
        self.assertEqual(self.version_1.tag, "eva-01")
        self.assertEqual(self.version_1.branch, "operation-yashima")

        self.version_2.refresh_from_db()
        self.assertEqual(self.version_2.branch, "operation-yashima")

        branch_version.refresh_from_db()
        self.assertIsNone(branch_version.branch)

    def test_update_worker_version_unique_tag(self):
        self.worker_reco.versions.create(
            docker_image_iid="registry.fake.com/workers/worker:12345",
            state=WorkerVersionState.Available,
            version=12,
            tag="eva-01"
        )

        with self.assertNumQueries(3):
            response = self.client.patch(
                reverse("api:version-retrieve", kwargs={"pk": str(self.version_1.id)}),
                data={
                    "tag": "eva-01"
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"tag": ["A version already exists for this worker with this tag."]})

    def test_update_worker_version_empty_tag_branch(self):
        with self.assertNumQueries(2):
            response = self.client.patch(
                reverse("api:version-retrieve", kwargs={"pk": str(self.version_1.id)}),
                data={
                    "tag": "",
                    "branch": "",
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "tag": ["This field may not be blank."],
            "branch": ["This field may not be blank."],
        })

    def test_cannot_update_worker_version_revision_ignored(self):
        self.version_1.revision_url = "url"
        self.version_1.version = None
        self.version_1.state = WorkerVersionState.Created
        self.version_1.save()
        with self.assertNumQueries(4):
            response = self.client.patch(
                reverse("api:version-retrieve", kwargs={"pk": str(self.version_1.id)}),
                data={
                    "revision_url": "updated"
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["id"], str(self.version_1.id))
        self.version_1.refresh_from_db()
        # revision_url just gets ignored
        self.assertEqual(self.version_1.revision_url, "url")

    def test_cannot_update_worker_version_revision_url_ignored(self):
        with self.assertNumQueries(4):
            response = self.client.patch(
                reverse("api:version-retrieve", kwargs={"pk": str(self.version_1.id)}),
                data={
                    "revision_url": "https://gitlab.com/NERV/eva/commit/eva-01"
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            # revision_url just gets ignored
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["id"], str(self.version_1.id))
        self.version_1.refresh_from_db()
        self.assertEqual(self.version_1.revision_url, None)

    def test_update_version_available_requires_docker_image(self):
        self.version_1.state = WorkerVersionState.Created
        self.version_1.docker_image_iid = None
        self.version_1.save()

        response = self.client.patch(
            reverse("api:version-retrieve", kwargs={"pk": str(self.version_1.id)}),
            data={"state": "available"},
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "state": [
                "A worker version must have a `docker_image_iid` set when updated to available."
            ]
        })

    def test_property_slug_explicit(self):
        version_2 = WorkerVersion.objects.create(
            id="01234567-89ab-cdef-fedc-ba9876543210",
            worker=self.worker_reco,
            configuration={"test": "test1"},
            gpu_usage=FeatureUsage.Disabled,
            version=2
        )
        self.assertEqual(version_2.slug, "reco_012345")

    def test_property_docker_command_empty(self):
        self.assertEqual(self.version_1.docker_command, None)

    def test_property_docker_command_not_empty(self):
        self.version_1.configuration = {"docker": {"command": "mysupercommand"}}
        self.assertEqual(self.version_1.docker_command, "mysupercommand")
        self.version_1.configuration = {"test": "test1"}

    def test_property_docker_shm_size(self):
        self.version_1.configuration = {"docker": {"shm_size": "32b"}}
        self.assertEqual(self.version_1.docker_shm_size, "32b")

    def test_corpus_worker_version_no_login_public(self):
        self.assertTrue(self.corpus.public)
        # Use multiple versions to ensure we don't get duplicated queries
        corpus_worker_version_1 = self.corpus.worker_version_cache.create(worker_version=self.version_1)
        corpus_worker_version_2 = self.corpus.worker_version_cache.create(worker_version=self.version_2)

        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:corpus-versions", kwargs={"pk": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 2,
            "number": 1,
            "previous": None,
            "next": None,
            "results": [{
                "id": str(corpus_worker_version_2.id),
                "worker_configuration": None,
                "model_version": None,
                "worker_version": {
                    "id": str(self.version_2.id),
                    "configuration": {"test": 42},
                    "revision_url": None,
                    "version": 1,
                    "tag": None,
                    "branch": None,
                    "gpu_usage": "disabled",
                    "model_usage": FeatureUsage.Disabled.value,
                    "docker_image_iid": self.version_2.docker_image_iid,
                    "state": "available",
                    "worker": {
                        "id": str(self.worker_dla.id),
                        "name": "Document layout analyser",
                        "type": "dla",
                        "slug": "dla",
                        "description": "",
                        "archived": False,
                        "repository_url": None,
                    },
                    "created": "2020-02-02T01:23:45.678000Z",
                },
            }, {
                "id": str(corpus_worker_version_1.id),
                "worker_configuration": None,
                "model_version": None,
                "worker_version": {
                    "id": str(self.version_1.id),
                    "configuration": {"test": 42},
                    "revision_url": None,
                    "version": 1,
                    "tag": None,
                    "branch": None,
                    "gpu_usage": "disabled",
                    "model_usage": FeatureUsage.Disabled.value,
                    "docker_image_iid": self.version_1.docker_image_iid,
                    "state": "available",
                    "worker": {
                        "id": str(self.worker_reco.id),
                        "name": "Recognizer",
                        "type": "recognizer",
                        "slug": "reco",
                        "description": "",
                        "archived": False,
                        "repository_url": None,
                    },
                    "created": "2020-02-02T01:23:45.678000Z",
                },
            }]
        })

    @patch("arkindex.users.managers.BaseACLManager.filter_rights", return_value=Corpus.objects.filter(public=True))
    def test_corpus_worker_version_no_login_private(self, filter_rights_mock):
        self.corpus.public = False
        self.corpus.save()
        self.corpus.worker_versions.set([self.version_1, self.version_2])

        with self.assertNumQueries(1):
            response = self.client.get(reverse("api:corpus-versions", kwargs={"pk": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertDictEqual(response.json(), {"detail": "No Corpus matches the given query."})

    def test_corpus_worker_version_not_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        # Use multiple versions to ensure we don't get duplicated queries
        corpus_worker_version_1 = self.corpus.worker_version_cache.create(worker_version=self.version_1)
        corpus_worker_version_2 = self.corpus.worker_version_cache.create(worker_version=self.version_2)

        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:corpus-versions", kwargs={"pk": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 2,
            "number": 1,
            "previous": None,
            "next": None,
            "results": [{
                "id": str(corpus_worker_version_2.id),
                "worker_configuration": None,
                "model_version": None,
                "worker_version": {
                    "id": str(self.version_2.id),
                    "configuration": {"test": 42},
                    "revision_url": None,
                    "version": 1,
                    "tag": None,
                    "branch": None,
                    "gpu_usage": "disabled",
                    "model_usage": FeatureUsage.Disabled.value,
                    "docker_image_iid": self.version_2.docker_image_iid,
                    "state": "available",
                    "worker": {
                        "id": str(self.worker_dla.id),
                        "name": "Document layout analyser",
                        "type": "dla",
                        "slug": "dla",
                        "description": "",
                        "archived": False,
                        "repository_url": None,
                    },
                    "created": "2020-02-02T01:23:45.678000Z",
                },
            }, {
                "id": str(corpus_worker_version_1.id),
                "worker_configuration": None,
                "model_version": None,
                "worker_version": {
                    "id": str(self.version_1.id),
                    "configuration": {"test": 42},
                    "revision_url": None,
                    "version": 1,
                    "tag": None,
                    "branch": None,
                    "gpu_usage": "disabled",
                    "model_usage": FeatureUsage.Disabled.value,
                    "docker_image_iid": self.version_2.docker_image_iid,
                    "state": "available",
                    "worker": {
                        "id": str(self.worker_reco.id),
                        "name": "Recognizer",
                        "type": "recognizer",
                        "slug": "reco",
                        "description": "",
                        "archived": False,
                        "repository_url": None,
                    },
                    "created": "2020-02-02T01:23:45.678000Z",
                },
            }]
        })

    def test_corpus_worker_version_list(self):
        self.client.force_login(self.user)
        # Use multiple versions with a model and worker configuration to ensure we don't get duplicated queries
        conf_1 = WorkerConfiguration.objects.create(name="test", configuration={"num": 1337}, worker=self.worker_reco)
        corpus_worker_version_1 = CorpusWorkerVersion.objects.create(
            corpus=self.corpus,
            worker_version=self.version_1,
            model_version=self.model_version,
            worker_configuration=conf_1,
        )
        conf_2 = WorkerConfiguration.objects.create(name="test 2", configuration={"value": "leet"}, worker=self.worker_dla)
        corpus_worker_version_2 = CorpusWorkerVersion.objects.create(
            corpus=self.corpus,
            worker_version=self.version_2,
            model_version=self.model_version,
            worker_configuration=conf_2,
        )

        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:corpus-versions", kwargs={"pk": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 2,
            "number": 1,
            "previous": None,
            "next": None,
            "results": [{
                "id": str(corpus_worker_version_2.id),
                "worker_version": {
                    "id": str(self.version_2.id),
                    "configuration": {"test": 42},
                    "revision_url": None,
                    "version": 1,
                    "tag": None,
                    "branch": None,
                    "gpu_usage": "disabled",
                    "model_usage": FeatureUsage.Disabled.value,
                    "docker_image_iid": self.version_2.docker_image_iid,
                    "state": "available",
                    "worker": {
                        "id": str(self.worker_dla.id),
                        "name": "Document layout analyser",
                        "type": "dla",
                        "slug": "dla",
                        "description": "",
                        "archived": False,
                        "repository_url": None,
                    },
                    "created": "2020-02-02T01:23:45.678000Z",
                },
                "worker_configuration": {
                    "id": str(conf_2.id),
                    "archived": False,
                    "name": "test 2",
                    "configuration": {"value": "leet"},
                },
                "model_version": {
                    "id": str(self.model_version.id),
                    "configuration": {},
                    "model": {
                        "id": str(self.model.id),
                        "name": "Generic model"
                    },
                    "size": 1337,
                    "state": ModelVersionState.Available.value,
                    "tag": None,
                },
            }, {
                "id": str(corpus_worker_version_1.id),
                "worker_version": {
                    "id": str(self.version_1.id),
                    "configuration": {"test": 42},
                    "revision_url": None,
                    "version": 1,
                    "tag": None,
                    "branch": None,
                    "gpu_usage": "disabled",
                    "model_usage": FeatureUsage.Disabled.value,
                    "docker_image_iid": self.version_1.docker_image_iid,
                    "state": "available",
                    "worker": {
                        "id": str(self.worker_reco.id),
                        "name": "Recognizer",
                        "type": "recognizer",
                        "slug": "reco",
                        "description": "",
                        "archived": False,
                        "repository_url": None,
                    },
                    "created": "2020-02-02T01:23:45.678000Z",
                },
                "worker_configuration": {
                    "id": str(conf_1.id),
                    "archived": False,
                    "name": "test",
                    "configuration": {"num": 1337},
                },
                "model_version": {
                    "id": str(self.model_version.id),
                    "configuration": {},
                    "model": {
                        "id": str(self.model.id),
                        "name": "Generic model"
                    },
                    "size": 1337,
                    "state": ModelVersionState.Available.value,
                    "tag": None,
                },
            }],
        })

    def test_corpus_worker_version_no_right(self):
        self.user.rights.all().delete()
        self.client.force_login(self.user)
        corpus_worker_version = self.corpus.worker_version_cache.create(worker_version=self.version_1)

        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:corpus-versions", kwargs={"pk": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 1,
            "number": 1,
            "previous": None,
            "next": None,
            "results": [{
                "id": str(corpus_worker_version.id),
                "worker_configuration": None,
                "model_version": None,
                "worker_version": {
                    "id": str(self.version_1.id),
                    "configuration": {"test": 42},
                    "revision_url": None,
                    "version": 1,
                    "tag": None,
                    "branch": None,
                    "gpu_usage": "disabled",
                    "model_usage": FeatureUsage.Disabled.value,
                    "docker_image_iid": self.version_1.docker_image_iid,
                    "state": "available",
                    "worker": {
                        "id": str(self.worker_reco.id),
                        "name": "Recognizer",
                        "type": "recognizer",
                        "slug": "reco",
                        "description": "",
                        "archived": False,
                        "repository_url": None,
                    },
                    "created": "2020-02-02T01:23:45.678000Z",
                }
            }]
        })
